import 'dart:async';
import 'dart:io';

import 'package:logging/logging.dart';
import 'package:selene/selene.dart';

void main() {
  devBot();
}

Future devBot() async {
  var client = new Client();
  Logger.root.onRecord.listen((LogRecord record) {
    print('[${record.loggerName}] ${record.level.name}: ${record.time}: ${record
        .message}');
  });

  client.dispatcher.onMessageReceive.listen((Message msg) async {
    if (msg.content == '!cache') {
      if (msg.channel is GuildChannel) {
        var guildChannel = msg.channel as GuildTextChannel;
        print('Getting cache for guild ' + guildChannel.guild.name);
        var guild = guildChannel.guild;
        var sb = new StringBuffer();
        var gc = guild.channels.values;
        await Future.forEach(gc, (GuildChannel guildChannel) async {
          if (guildChannel is CategoryChannel) {
            sb.writeln('Category ${guildChannel.name}');
            await Future.forEach(guildChannel.channels, (GuildChannel child) {
              sb.writeln('      ${child} (`${child.runtimeType}`)');
            });
          } else if (guildChannel.category == null) {
            sb.writeln('${guildChannel}');
          }
        });
        await guildChannel.sendMessage(
            content: (sb.toString().isEmpty ? 'Nothing.' : sb.toString()));
      } else if (msg.channel is DMChannel) {
        await msg.channel.sendMessage(content: 'Hello DM!');
      }
    }
    if (msg.content == '!throw') {
      throw new UnsupportedError('Test Exception');
    }
    if (msg.content == '!delete') {
      var newMsg = await msg.channel.sendMessage(
          content: "I am going to delete this message in 3 seconds.");
      await new Future.delayed(new Duration(seconds: 3), () async {
        await newMsg.delete();
      });
    }
    if (msg.content.startsWith('!guild')) {
      var methods = {
        'info': (args) async {
          await msg.channel.sendMessage(
              content: (msg.channel as GuildTextChannel).guild.toString());
        },
        'create': (args) async {
          print('Creating guild with name ${args[0]}');
          var newGuild = await client.createGuild(
              args[0],
              client.state.voiceRegions.values
                  .firstWhere((reg) => reg.id == args[1], orElse: () {
                return null;
              }),
              null,
              0,
              0,
              0);
          await msg.channel.sendMessage(
              content: 'Created ${newGuild.name} with ID ${newGuild.id}');
        },
        'delete': (args) async {
          var guild = client.state.guilds.values
              .firstWhere((g) => g.id == int.parse(args[0]), orElse: () {
            throw 'Unknown guild';
          });
          await guild.delete();
        },
        'modify-name': (args) async {
          var guild = client.state.guilds.values
              .firstWhere((g) => g.id == int.parse(args[0]), orElse: () {
            throw 'Unknown guild';
          });
          await guild.modify(name: args[1]);
          await msg.channel.sendMessage(
              content: 'Changed guild name. New name: ${guild.name}');
        }
      };
      var args = msg.content.split(" ");
      if (args[0] != '!guild') return;
      if (!methods.keys.contains(args[1])) return;
      args.removeAt(0);
      await methods[args.removeAt(0)](args);
    }
    if (msg.content == '!edit') {
      var oldMsg = await msg.channel.sendMessage(
          content: "This is the old message. Updating in 3 seconds.");
      await new Future.delayed(new Duration(seconds: 3), () async {
        await oldMsg.edit(content: 'This is the new message.');
      });
    }
    if (msg.content == '!self') {
      var self = client.state.currentUser;
      var embed = <String, dynamic>{
        'author': {'name': self.username, 'icon_url': self.getAvatarUrl()},
        'description': 'I am the Selene developer test bot.'
      };
      await msg.channel.sendMessage(embed: embed);
    }
    if (msg.content == '!guild') {
      var message = msg.channel as GuildTextChannel;
      var guild = message.guild;
      var embed = {
        'title': 'Guild ${guild.name}',
        'description': 'Guild properties overview',
        'fields': [
          {'name': 'ID', 'value': guild.id ?? 'blank', 'inline': true},
          {
            'name': 'Channel Length',
            'value': guild.channels.length,
            'inline': true
          },
          {
            'name': 'AFK Channel ID',
            'value': guild.afkChannelId ?? 'none',
            'inline': true
          },
          {
            'name': 'AFK Timeout',
            'value': guild.afkTimeout ?? 'none',
            'inline': true
          },
          {
            'name': 'Roles',
            'value': (guild.roles ?? [])
                .map((GuildRole role) => role.name)
                .join(", "),
            'inline': true
          },
          {'name': 'Voice Region', 'value': guild.voiceRegion, 'inline': true}
        ]
      };

      await msg.channel.sendMessage(embed: embed);
    }
    if (msg.content == '!ratelimit') {
      // Attempt to invoke message send rate limit
      await Future.forEach(
          new Iterable.generate(15, (ix) {
            return ix;
          }), (i) async {
        await msg.channel.sendMessage(content: "Rate limiting - message ${i}");
      });
    }
    if (msg.content == '!snowflake') {
      var snowflake = msg.snowflakeId;
      await msg.channel.sendMessage(
          content: 'Your message was sent at ${snowflake.timestamp
              .toString()}, on the worker ${snowflake
              .workerId}, on process ${snowflake
              .processId}, with an internal ID of ${snowflake.idOnProcess}.');
    }
    if (msg.content == '!member') {
      var member = msg.member;
      var embed = <String, dynamic>{
        'title': 'Info for ${member.nickname ?? member.user.username}#${member
            .user.discriminator}',
        'description': 'Has ${member.roles.length} roles',
        'fields': [
          {
            'name': 'Roles',
            'value': emptyReplace(
                member.roles.map((role) => role.name).join(", "), 'None'),
            'inline': true,
          },
          {
            'name': 'Deafened/Muted',
            'value': '${member.isDeafened}/${member.isMuted}',
            'inline': true
          },
          {
            'name': 'Has Nickname',
            'value': '${member.nickname != null
                ? 'Yes (${member.nickname})'
                : 'No'}.',
            'inline': true
          }
        ]
      };
      await msg.channel.sendMessage(embed: embed);
    }
    if (msg.content == '!category') {
      await msg.channel.sendMessage(
          content: 'This channel\'s category is ${(msg
              .channel as GuildTextChannel).category.name}.');
    }
    if (msg.content == '!channel') {
      if (msg.channel.type == ChannelType.DM) {
        var dm = msg.channel as DMChannel;
        await msg.channel.sendMessage(content: dm.toString());
        return;
      }

      var channel = msg.channel as GuildTextChannel;
      var embed = <String, dynamic>{
        'title': 'Info for channel ${channel.name}',
        'description': 'Discord channel information',
        'fields': [
          {
            'name': 'Topic',
            'value': channel.topic ?? 'No topic',
            'inline': true
          }
        ]
      };

      await channel.sendMessage(embed: embed);
    }
    if (msg.content == '!dumpall') {
      String dumpUsers() {
        return "**Users**\n" +
            client.state.users.values
                .map((User user) => user.toString())
                .join("\n");
      }

      String dumpGuilds() {
        return "**Guilds**\n" +
            client.state.guilds.values
                .map((Guild guild) => guild.toString())
                .join("\n");
      }

      String dumpDms() {
        return "**DMs**\n" +
            client.state.dmChannels.values
                .map((DMChannel channel) => channel.toString())
                .join("\n");
      }

      String dumpMembers() {
        return "**Members**\n" +
            client.state.members.values
                .map((GuildMember member) => member.toString())
                .join("\n");
      }

      await msg.channel.sendMessage(
          content: dumpUsers() +
              "\n\n" +
              dumpGuilds() +
              "\n\n" +
              dumpDms() +
              "\n\n" +
              dumpMembers());
    }
    if (msg.content == '!members') {
      await msg.channel.sendMessage(
          content: (msg.channel as GuildTextChannel)
              .guild
              .members
              .values
              .map((member) => member.user.username)
              .join(", "));
    }
    if (msg.content == '!messages') {
      var messageString = msg.channel.messages
          .map((id, msg) =>
      new MapEntry(id, "${msg.author.username}: ${msg.content}"))
          .values
          .join("\n");
      await msg.channel.sendMessage(content: messageString);
    }
    if (msg.content == '!rename') {
      await msg.channel.modify(
          name: 'newchannel2',
          topic: 'Renamed channel topic',
          category: client.state.resolveChannel(462559251420938240));
      await msg.channel.sendMessage(content: 'Done');
    }
  });
  client.dispatcher.onReady.listen((_) {
    new Logger('Selene-Test').info('Client is ready');
  });

  client.login(Platform.environment['SELENE_TOKEN']);
  await client.start();
}
